@@ -36,6 +36,8 @@ class ConsoleModule(mp_module.MPModule):
         self.last_sys_status_errors_announce = 0
         self.user_added = {}
         self.safety_on = False
+        self.unload_check_interval = 5 # seconds
+        self.last_unload_check_time = time.time()
         self.add_command('console', self.cmd_console, "console module", ['add','list','remove'])
         mpstate.console = wxconsole.MessageConsole(title='Console')
 
@@ -86,7 +88,8 @@ class ConsoleModule(mp_module.MPModule):
             self.menu = MPMenuTop([])
             self.add_menu(MPMenuSubMenu('MAVProxy',
                                         items=[MPMenuItem('Settings', 'Settings', 'menuSettings'),
-                                               MPMenuItem('Map', 'Load Map', '# module load map')]))
+                                               MPMenuItem('Show Map', 'Load Map', '# module load map'),
+                                               MPMenuItem('Show HUD', 'Load HUD', '# module load horizon')]))
             self.vehicle_menu = MPMenuSubMenu('Vehicle', items=[])
             self.add_menu(self.vehicle_menu)
 
@@ -199,6 +202,8 @@ class ConsoleModule(mp_module.MPModule):
             return "Tracker"
         if hb.type == mavutil.mavlink.MAV_TYPE_AIRSHIP:
             return "Blimp"
+        elif hb.type == mavutil.mavlink.MAV_TYPE_ADSB:
+            return "ADSB"
         return "UNKNOWN(%u)" % hb.type
 
     def component_type_string(self, hb):
@@ -209,6 +214,8 @@ class ConsoleModule(mp_module.MPModule):
             return "Gimbal"
         elif hb.type == mavutil.mavlink.MAV_TYPE_ONBOARD_CONTROLLER:
             return "CC"
+        elif hb.type == mavutil.mavlink.MAV_TYPE_ADSB:
+            return "ADSB"
         elif hb.type == mavutil.mavlink.MAV_TYPE_GENERIC:
             return "Generic"
         return self.vehicle_type_string(hb)
@@ -481,7 +488,7 @@ class ConsoleModule(mp_module.MPModule):
                 status += 'O2'
             self.console.set_status('PWR', status, fg=fg)
             self.console.set_status('Srv', 'Srv %.2f' % (msg.Vservo*0.001), fg='green')
-        elif type == 'HEARTBEAT':
+        elif type in ['HEARTBEAT', 'HIGH_LATENCY2']:
             fmode = master.flightmode
             if self.settings.vehicle_name:
                 fmode = self.settings.vehicle_name + ':' + fmode
@@ -590,7 +597,7 @@ class ConsoleModule(mp_module.MPModule):
             rec, tot = self.module('param').param_status()
             self.console.set_status('Params', 'Param %u/%u' % (rec,tot))
             
-        elif type == 'HIGH_LATENCY2':
+        if type == 'HIGH_LATENCY2':
             self.console.set_status('WPDist', 'Distance %s' % self.dist_string(msg.target_distance * 10))
             # The -180 here for for consistency with NAV_CONTROLLER_OUTPUT (-180->180), whereas HIGH_LATENCY2 is (0->360)
             self.console.set_status('WPBearing', 'Bearing %u' % ((msg.target_heading * 2) - 180))
@@ -653,6 +660,13 @@ class ConsoleModule(mp_module.MPModule):
                     print(ex)
                     pass
 
+    def idle_task(self):
+        now = time.time()
+        if self.last_unload_check_time + self.unload_check_interval < now:
+            self.last_unload_check_time = now
+            if not self.console.is_alive():
+                self.needs_unloading = True
+
 def init(mpstate):
     '''initialise module'''
     return ConsoleModule(mpstate)
