@@ -372,7 +372,7 @@ class ParamState:
                 print("Usage: param set PARMNAME VALUE")
                 return
             if len(args) == 2:
-                self.mav_param.show(args[1])
+                self.param_show(args[1], self.mpstate.settings.param_docs)
                 return
             param = args[1]
             value = args[2]
@@ -436,7 +436,7 @@ class ParamState:
         elif args[0] == "set_xml_filepath":
             self.param_set_xml_filepath(args[1:])
         elif args[0] == "show":
-            verbose = False
+            verbose = self.mpstate.settings.param_docs
             if len(args) > 1:
                 pattern = args[1]
                 if len(args) > 2 and args[2] == '-v':
@@ -576,7 +576,10 @@ class ParamModule(mp_module.MPModule):
                                          MPMenuItem('Save', 'Save', '# param save ',
                                                     handler=MPMenuCallFileDialog(flags=('save', 'overwrite_prompt'),
                                                                                  title='Param Save',
-                                                                                 wildcard='ParmFiles(*.parm,*.param)|*.parm;*.param'))])
+                                                                                 wildcard='ParmFiles(*.parm,*.param)|*.parm;*.param')),
+                                         MPMenuItem('FTP', 'FTP', '# param ftp'),
+                                         MPMenuItem('Update Metadata', 'Update Metadata', '# param download'),
+                                             ])
 
     def get_component_id_list(self, system_id):
         '''get list of component IDs with parameters for a given system ID'''
@@ -637,9 +640,12 @@ class ParamModule(mp_module.MPModule):
         self.pstate[sysid].vehicle_name = self.vehicle_name
         self.pstate[sysid].param_help.vehicle_name = self.vehicle_name
         self.pstate[sysid].fetch_check(self.master)
-        if self.module('console') is not None and not self.menu_added_console:
-            self.menu_added_console = True
-            self.module('console').add_menu(self.menu)
+        if self.module('console') is not None:
+            if not self.menu_added_console:
+                self.menu_added_console = True
+                self.module('console').add_menu(self.menu)
+        else:
+            self.menu_added_console = False
 
     def cmd_param(self, args):
         '''control parameters'''
